unit MMMainFrm;
{------------------------------------------------------------------------------}
(*
  This demo requires TRichView 17.3 or newer.
  

  Mail merging application supporting the following field values:
  - multiline text;
  - pictures;
  - tables
  - RVF document
  Fields are any strings inside {}.
  Supported fields:
  Text:
    {NAME}, {ADDRESS1}, {ADDRESS2}, {CITY}, {ZIP}, {COUNTRY}, {FULL_ADDRESS}.
    {FULL_ADDRESS} is multiline.
  Picture:
    {PIC#}, where # is a picture size (e.g {PIC100}). The default size is 24.
  Table:
    {TABLE#}, where # is a table size (count of rows and columns). The default size is 3.
  RVF:
    {COMMENTS}
  This demo generates report for all records as a single document (separated by
  page breaks).

  v1.0 - initial release
  v1.1 - RVF fields are added
  v1.2 - compatibility with TRichView 17.3. Tab characters are allowed in
    text values. Unicode text.
*)
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, RVScroll, RichView, StdCtrls, RVStyle, RVTable, CRVData, RVItem,
  RVEdit, RVUni, RVTypes, RVInsertItems;

{$I RV_Defs.inc}

type

  { Information about a person, used for text fields }
  TPerson = class (TCollectionItem)
    public
      Name, Address1, Address2, Zip, City, Country, Comments: String;
  end;

  { type of field value }
  TMMFieldType = (mmftText, mmftPicture, mmftTable, mmftRVF);


  TForm1 = class(TForm)
    rvOutput: TRichView;
    rvsOutput: TRVStyle;
    rvsTemplate: TRVStyle;
    Label1: TLabel;
    Label2: TLabel;
    rve: TRichViewEdit;
    Button1: TButton;
    Label3: TLabel;
    Label4: TLabel;
    rveComments: TRichViewEdit;
    rvsComments: TRVStyle;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    { Private declarations }
    Persons: TCollection;
    Index: Integer;
    procedure LoadPersons;
    procedure CreateTemplate;
    procedure InitComments;
    function GetFieldTyle(const FieldName: TRVUnicodeString): TMMFieldType;
    function GetFieldValueFromDatabase(const FieldName: TRVUnicodeString): TRVUnicodeString;
    function GetPicture(const FieldName: TRVUnicodeString): TGraphic;
    function GetTable(RVData: TCustomRVData; const FieldName: TRVUnicodeString): TRVTableItemInfo;
    procedure GetRVFFieldValue(const FieldName: TRVUnicodeString; Stream: TStream);


    procedure FillFields(RVData: TCustomRVData; StartItemNo: Integer);
    procedure InsertField(RVData: TCustomRVData;
      var ItemNo, Pos: Integer; const FieldName: TRVUnicodeString);
    procedure InsertItem(RVData: TCustomRVData; var ItemNo, Pos: Integer;
      const Item: TCustomRVItemInfo);
    procedure InsertRVF(RVData: TCustomRVData; var ItemNo: Integer;
      var Pos: Integer; Stream: TStream);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses
  RVInsertMultilineText; // <- included in this project

{$R *.dfm}

{------------------------------------------------------------------------------}
{ This procedure loads file CUSTOMERS.TXT into Persons.
  In real applications data can be stored in some database }
procedure TForm1.LoadPersons;
var F: TextFile;
    i, Count: Integer;
    Person: TPerson;
begin
  AssignFile(F, ExtractFilePath(Application.ExeName)+'Persons.txt');
  Reset(F);
  try
    Readln(F, Count);
    for i := 0 to Count-1 do
    begin
      Person := Persons.Add as TPerson;
      Readln(F, Person.Name);
      Readln(F, Person.Address1);
      Readln(F, Person.Address2);
      Readln(F, Person.City);
      Readln(F, Person.Zip);
      Readln(F, Person.Country);
    end;
  finally
    CloseFile(F);
  end;
  Label3.Caption := IntToStr(Persons.Count) + ' personal records in the database';
end;
{------------------------------------------------------------------------------}
{ Creating a sample template. You can edit it }
procedure TForm1.CreateTemplate;
begin
   rve.Clear;
   rve.AddNL('To {NAME}', 0, 0);
   rve.AddNL('{FULL_ADDRESS}', 0, 0);
   rve.AddNL('Letter Sample', 1, 1);
   rve.AddNL('Hello {NAME}', 0, 0);
   rve.AddNL('', 0, 0);
   rve.AddNL('You can find a yellow picture {PIC20} and a table {TABLE5} in this message.', 0, 0);
   rve.AddNL('', 0, 0);
   rve.AddNL('With best regards,', 0, 0);
   rve.AddNL('your template generator.', 0, 0);
   rve.AddNL('PS: {COMMENTS}', 0, 0);
   rve.AddNL('', 0, 0);
   rve.AddNL('', 0, 0);
   rve.AddNL('', 0, 0);
   rve.Format;
end;
{------------------------------------------------------------------------------}
{
  Initializing comments
}
procedure TForm1.InitComments;
begin
  rveComments.Clear;
  rveComments.AddNL('I can use fonts', 0, 0);
  rveComments.Format;
end;
{------------------------------------------------------------------------------}
{
  This function returns the field type by the field name
}
function TForm1.GetFieldTyle(const FieldName: TRVUnicodeString): TMMFieldType;
begin
  if Copy(FieldName, 1, 3)='PIC' then
    Result := mmftPicture
  else if Copy(FieldName, 1, 5)='TABLE' then
    Result := mmftTable
  else if FieldName='COMMENTS' then
    Result := mmftRVF
  else
    Result := mmftText;
end;
{------------------------------------------------------------------------------}
{
   Fills stream with RVF document by the field code.
 }
procedure TForm1.GetRVFFieldValue(const FieldName: TRVUnicodeString; Stream: TStream);
begin
  Stream.Size := 0;
  if FieldName='COMMENTS' then
    rveComments.SaveRVFToStream(Stream, False);
end;
{------------------------------------------------------------------------------}
{
  Creates a picture by the field code.
  This example creates a yellow square of the specified size.
  In your application, if the field code is invalid, you can return nil.
}
function TForm1.GetPicture(const FieldName: TRVUnicodeString): TGraphic;
var Size: Integer;
begin
  Size := StrToIntDef(Copy(FieldName, 4, Length(FieldName)), 24);
  Result := TBitmap.Create;
  Result.Width := Size;
  Result.Height := Size;
  TBitmap(Result).Canvas.Brush.Color := clYellow;
  TBitmap(Result).Canvas.FillRect(Rect(0,0,Size,Size));
end;
{------------------------------------------------------------------------------}
{
  Creates a table by the field code.
  This example creates table of the specified size with "Sample Table" text.
  In your application, if the field code is invalid, you can return nil.
}
function TForm1.GetTable(RVData: TCustomRVData;
  const FieldName: TRVUnicodeString): TRVTableItemInfo;
var Size: Integer;
begin
  Size := StrToIntDef(Copy(FieldName, 6, Length(FieldName)), 3);
  Result := TRVTableItemInfo.CreateEx(Size, Size, RVData);
  Result.BorderWidth := 3;
  Result.CellBorderWidth := 1;
  Result.BestWidth := -100;
  Result.MergeCells(0, 0, Result.ColCount, 1, True);
  Result.Cells[0,0].Clear();
  Result.Cells[0,0].AddNL('Sample Table', 0, 0);
end;
{------------------------------------------------------------------------------}
{
  This function returns the current text field value by the field name.
}
function TForm1.GetFieldValueFromDatabase(
  const FieldName: TRVUnicodeString): TRVUnicodeString;
var Person: TPerson;
begin
  Person := Persons.Items[Index] as TPerson;
  if FieldName='NAME' then
    Result := Person.Name
  else if FieldName='ADDRESS1' then
    Result := Person.Address1
  else if FieldName='ADDRESS2' then
    Result := Person.Address2
  else if FieldName='CITY' then
    Result := Person.City
  else if FieldName='ZIP' then
    Result := Person.ZIP
  else if FieldName='COUNTRY' then
    Result := Person.Country
  else if FieldName='FULL_ADDRESS' then
    Result := Person.Address1+#13+Person.Address2+#13+Person.City+' '+
      Person.ZIP+#13+Person.Country
  else
    Result := '{error: illegal field code}';
end;
{------------------------------------------------------------------------------}
procedure TForm1.FormCreate(Sender: TObject);
begin
  Persons := TCollection.Create(TPerson);
  LoadPersons;
  CreateTemplate;
  InitComments;
end;
{------------------------------------------------------------------------------}
procedure TForm1.FormDestroy(Sender: TObject);
begin
  Persons.Free;
  Persons := nil;
end;
{------------------------------------------------------------------------------}
{ Generating output }
procedure TForm1.Button1Click(Sender: TObject);
var Stream: TStream;
    ItemCount: Integer;
begin
  Index := 0; // index of the current person
  ItemCount := 0;
  rvOutput.Clear;
  while Index<Persons.Count do
  begin
    // copying rve to the end of rvOutput
    Stream := TMemoryStream.Create;
    rve.SaveRVFToStream(Stream, False);
    Stream.Position := 0;
    rvOutput.InsertRVFFromStream(Stream, ItemCount);
    Stream.Free;
    // processing the last added copy of template
    if rvOutput.ItemCount>ItemCount then
    begin
      // starting from new page
      if ItemCount>0 then
        rvOutput.PageBreaksBeforeItems[ItemCount] := True;
      // replacing field codes
      FillFields(rvOutput.RVData, ItemCount);
    end;
    ItemCount := rvOutput.ItemCount;
    inc(Index);
  end;
  rvOutput.Format;
end;
{==============================================================================}
{ The code below this line can be copied your application unchanged }
{------------------------------------------------------------------------------}
(*
  This function iterates through all items in RVData starting from StartItemNo,
  searches for strings inside {}, calls InsertField()
  and replaces text between {} (including {}) with the text/picture/table.
  You can move this function to your application unchanged.
*)
procedure TForm1.FillFields(RVData: TCustomRVData; StartItemNo: Integer);
var
  i,j,r,c: Integer;
  table: TRVTableItemInfo;
  s, FieldName: TRVUnicodeString;
  FieldStart: Integer;
begin
  i := StartItemNo;
  while i<RVData.ItemCount do
  begin
    if RVData.GetItem(i) is TRVTableItemInfo then
    begin // recursive processing of tables
      table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to table.RowCount-1 do
        for c := 0 to table.ColCount-1 do
          if table.Cells[r,c]<>nil then
            FillFields(table.Cells[r,c].GetRVData, 0);
      end
    else if RVData.GetItemStyle(i)>=0 then
    begin
      s := RVData.GetItemTextW(i);
      FieldStart := 0;
      j := 1;
      while j<=Length(s) do
      begin
        if s[j]='{' then // may be this is a beginning of field code
          FieldStart := j
        else if (s[j]='}') and (FieldStart>0) then
        begin // field code is found
          FieldName :=  Copy(s, FieldStart+1, j-FieldStart-1);
          // deleting field code
          Delete(s, FieldStart, j-FieldStart+1);
          RVData.SetItemTextW(i, s);
          // inserting field value
          j := FieldStart;
          InsertField(RVData, i, j, FieldName);
          if RVData.GetItemStyle(i)<0 then
            s := ''
          else
            s := RVData.GetItemTextW(i);
          FieldStart := 0;
          continue;
        end;
        inc(j);
      end;
    end;
    inc(i);
  end;
end;
{------------------------------------------------------------------------------}
{
  This function inserts a field value in the position (RVData, ItemNo, Pos).
  After inserting, (RVData, ItemNo, Pos) is a position after the inserted item.
  You can move this function to your application unchanged.
}
procedure TForm1.InsertField(RVData: TCustomRVData;
  var ItemNo, Pos: Integer; const FieldName: TRVUnicodeString);
var Item: TCustomRVItemInfo;
    Graphic: TGraphic;
    Stream: TMemoryStream;
begin
  case GetFieldTyle(FieldName) of
    mmftText:
      InsertMultilineText(RVData, ItemNo, Pos, GetFieldValueFromDatabase(FieldName));
    mmftRVF:
      begin
        Stream := TMemoryStream.Create;
        try
          GetRVFFieldValue(FieldName, Stream);
          InsertRVF(RVData, ItemNo, Pos, Stream);
        finally
          Stream.Free;
        end;
      end;
    mmftPicture:
      begin
        Graphic := GetPicture(FieldName);
        if Graphic=nil then
        begin
          InsertMultilineText(RVData, ItemNo, Pos, '{error: invalid picture}');
          exit;
        end;
        Item := TRVGraphicItemInfo.CreateEx(RVData, Graphic, rvvaBaseline);
        InsertItem(RVData, ItemNo, Pos, Item);
      end;
    mmftTable:
      begin
        Item := GetTable(RVData, FieldName);
        if Item=nil then
        begin
          InsertMultilineText(RVData, ItemNo, Pos, '{error: invalid table}');
          exit;
        end;
        InsertItem(RVData, ItemNo, Pos, Item);
      end;
  end;
end;
{------------------------------------------------------------------------------}
procedure TForm1.InsertRVF(RVData: TCustomRVData;
      var ItemNo: Integer; var Pos: Integer; Stream: TStream);
var
  rv: TRichView;
  Stream2: TMemoryStream;
  s1, s2: TRVUnicodeString;
  Color: TColor;
  ParaNo, j: Integer;
  BR, ContinuePara: Boolean;
  ItemsAdded: Integer;
begin
  // splitting the item at the place of insertion
  s1 := RVData.GetItemTextW(ItemNo);
  s2 := Copy(s1, Pos, Length(s1));
  s1 := Copy(s1, 1, Pos-1);
  ParaNo := RVData.GetItemPara(ItemNo);
  if s2<>'' then
  begin
    RVData.SetItemTextW(ItemNo, s1);
    RVInsertStringW(RVData, ItemNo + 1, s2, RVData.GetItemStyle(ItemNo),
      -1, RVData.GetItemTag(ItemNo));
  end;
  if s1='' then
  begin
    BR := RVData.GetItem(ItemNo).BR;
    ContinuePara := RVData.GetItem(ItemNo).SameAsPrev;
    RVData.DeleteItems(ItemNo, 1);
    dec(ItemNo);
  end
  else
  begin
    BR := False;
    ContinuePara := True;
  end;
  inc(ItemNo);
  // inserting RVF
  rv := TRichView.Create(nil);
  try
    rv.Style := RVData.GetRVStyle;
    rv.Options := RVData.Options;
    rv.RVFOptions := RVData.RootProperties.RVFOptions - [rvfoSaveTextStyles, rvfoSaveParaStyles];
    Stream.Position := 0;
    rv.InsertRVFFromStream(Stream, 0);
    Stream2 := TMemoryStream.Create;
    try
      rv.SaveRVFToStream(Stream2, False);
      Stream2.Position := 0;
      Color := clNone;
      ItemsAdded := RVData.ItemCount;
      RVData.InsertRVFFromStream(Stream2, ItemNo, Color, nil, nil, False, nil, nil);
      ItemsAdded := RVData.ItemCount-ItemsAdded;
    finally
      Stream2.Free;
    end;
  finally
    rv.Free;
  end;
  // making corrections
  for j := ItemNo to RVData.ItemCount-1 do
  begin
    if (ItemNo=j) then begin
      if RVData.GetItem(j).GetBoolValue(rvbpFullWidth) or
        (RVData.GetItemStyle(ItemNo)=rvsListMarker) then
      begin
        if (ItemNo>0) and (RVData.GetItemStyle(ItemNo-1)=rvsListMarker) then
        begin
          dec(ItemNo);
          RVData.DeleteItems(ItemNo, 1);
        end;
        break;
      end;
      RVData.GetItem(j).SameAsPrev := ContinuePara;
      if BR then
        RVData.GetItem(j).BR := True;
    end;
    if (j>ItemNo) and RVData.IsParaStart(j) then
      break;
    RVData.GetItem(j).ParaNo := ParaNo;
  end;
  inc(ItemNo, ItemsAdded-1);
  Pos := RVData.GetOffsAfterItem(ItemNo)
end;
{------------------------------------------------------------------------------}
{ Inserts item (for example, picture or table) in the position (RVData, ItemNo, Pos).
  After inserting, (RVData, ItemNo, Pos) is a position after the inserted item.
  This method uses some undocumented RichView functions. Just copy it to your
  application. }
procedure TForm1.InsertItem(RVData: TCustomRVData;
  var ItemNo, Pos: Integer; const Item: TCustomRVItemInfo);
var
  s1, s2, s: TRVUnicodeString;
  StyleNo, ParaNo: Integer;
  FullWidthItem: Boolean;
  Tag: TRVTag;
begin
  s1 := RVData.GetItemTextW(ItemNo);
  s2 := Copy(s1, Pos, Length(s1));
  s1 := Copy(s1, 1, Pos-1);
  ParaNo := RVData.GetItemPara(ItemNo);
  Tag := RVData.GetItemTag(ItemNo);
  Item.ParaNo := ParaNo;
  StyleNo := RVData.GetItemStyle(ItemNo);
  FullWidthItem := Item.GetBoolValue(rvbpFullWidth);
  Item.ParaNo := ParaNo;
  if s1<>'' then
  begin
   RVData.SetItemTextW(ItemNo, s1);
   if not FullWidthItem then
     Item.SameAsPrev := True;
   inc(ItemNo);
  end
  else
  begin
    if not FullWidthItem then
      Item.BR := RVData.GetItem(ItemNo).BR;
    Item.PageBreakBefore := RVData.GetItem(ItemNo).PageBreakBefore;
    RVData.DeleteItems(ItemNo, 1);
  end;
  s := '';
  Item.Inserting(RVData, s, False);
  RVData.Items.InsertObject(ItemNo, s, Item);
  Item.Inserted(RVData, ItemNo);
  Pos := 1;

  if s2<>'' then
  begin
    if not FullWidthItem then
      ParaNo := -1;
    RVInsertStringW(RVData, ItemNo+1, s2, StyleNo, ParaNo, Tag);
  end
  else if FullWidthItem and (RVData.ItemCount>ItemNo+1) then
    RVData.GetItem(ItemNo+1).SameAsPrev := False;
end;

end.
